% DEMO
% Blind Deconvolution in Multiframe Imaging Using
% Anisotropic Spatially-Adaptive Denoising
%
% This demo performs blind deconvoluion using multiple images which are
% blurred and noisy. The noise is assumed to be white Gaussian. The LPA-ICI
% technique is used as an adaptive to signal denoising and as a 
% regularizator.
%
% The files are prepared in Tampere University of Technology, Institute of
% Information Technology, 2006.
%
% REFERENCE
% Katkovnik V., D. Paliy, K. Egiazarian, and J. Astola, Frequency domain 
% blind deconvolution in multiframe imaging using anisotropic spatially-
% adaptive denoising, Proc. 14th European Signal Process. Conf., EUSIPCO 
% 2006, Florence, September 2006.
%
% Vladimir Katkovnik, Alessandro Foi, Dmitriy Paliy*
% *e-mail: dmitriy.paliy@tut.fi
%
% =========================================================================
%
% OBSERVATION MODEL:
% z(j) = y*v(j)+sigma(j)n
% z(j) : blurred noisy observations, j=1,...,L.
% y    : true image (assumed as unknown)
% v(j) : point-spread functions
% n    : gaussian white noise with standard deviation sigma(j) for every 
%        jth obsrvation z(j)
% *    : convolution
%
% The problem is to reconstruct y having only blurred and noisy 
% observations z(j)
%
% THE ALGORITHM'S MAIN OPTIONS:
%
% ITER_MAX = 20; - the maximal number of iterations
%
% COLOR_METHOD = 'OPP'; - the colorspace transformation method used if the 
%                images z(j) are 3 dimensional (color). Possible values 
%                'RGB', 'OPP', 'DCT', 'YUV', 'PCA'.
%                'OPP' - the opponent color space
%
% GammaSignal = [.9 .9 .9]; - the crucial threshold parameter for the 
%                ICI rule. Three values are for the 3 color channels. If 
%                the image is grayscale then only the first value is used.
%
% GammaParameterPSF = [1.4]; % this parameter always fixed and does not 
%                            % depend on the color channel
%
% h1_Im, h2_Im    - set of scales H used for the LPA kernels design for 
%                   IMAGE filtering by LPA-ICI
%
% h1_PSF, h2_PSF  - set of scales H used for the LPA kernels design for 
%                   PSFs filtering by LPA-ICI
%
% stepsize_v=.6;  - step sizes in iterative scheme, for v
% stepsize_y=.9;    and for y restoration
%
% deltaPSF=15;    - support size for PSF projection. The region out of
%                   deltaPSF x deltaPSF is set to be 0 on every iteration.
%
% Input.y         - is a true testing image
% Input.v_true{j} - blurring PSF
% Input.BSNR(j)   - level of in dB for jth observation z(j)
% [Input.shifts(j,1) Input.shifts(j,2)] - initial shifts for the simulation 
%                   of misaligned images
% IMPORTANT!!!!
% Number of frames (observations z(j)) is defined by the number of PSFs
% Input.v_true{j}, noise level Input.BSNR(j), and shifts 
% [Input.shifts(j,1) Input.shifts(j,2)] for each frame. If noise level
% (BSNR) and shifts are not defined for all PSFs Input.v_true{j} then
% default values are set to be BSNR=40dB and 0-shift.
%
% IMAGE_normalization = 'ON'; % 'ON', 'OFF'. normalize the image at every 
%                             iteration to be in [0,1]
%
% PSF_positivity = 'ON'; % 'ON', 'OFF'. positivity of PSFs is assumed to 
%                        be known a priory
% 
% PSF_normalization = 'ON'; % 'ON', 'OFF'. PSFs are normalized on every 
%                           iteration - 'ON'
%
% EXPERIMENT = 'cameraman'; % - grayscale 256x256
%              'fruits'; % - color RGB image 512x512
%              'lena' (grayscale 256x256),
%              'boats' (grayscale 512x512),
%              'text' (grayscale 256x256),
%              'testpat' (grayscale 256x256),
%
% series of the signal and PSFs estimates at every iteration are stored in 
% the variables y_est_series and v_est_series, respectively. use function
% function_EstimationShow to view them
%
% function_EstimationShow(y_est_series)
%
% Other parameters can be found in Section ALGORITHM PARAMETERS

clear all
close all

% ----------------------------------------------------------------------
% Algorithm's MAIN options
% ----------------------------------------------------------------------

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% LPA-ICI Settings: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
GammaSignal = [.9]; % for color images (fruits)
GammaParameterPSF = [1.5];   % Gamma for PSFs

% set of scales used for LPA for the IMAGE Input.y filtering
h1_Im=[1 3 5 7 11]; 
h2_Im=[1 1 1 2 2];

% set of scales used for LPA for the PSFs Input.v_true{j} filtering
h1_PSF=[1 2 3 5 11];
h2_PSF=[1 1 1 2 2];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% Experiment settings: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
EXPERIMENT = 'cameraman'; % 'cameraman' (grayscale 256x256),
                          % 'fruits' (color RGB),
                          % 'lena' (grayscale 256x256),
                          % 'boats' (grayscale 512x512),
                          % 'text' (grayscale 256x256),
                          % 'testpat' (grayscale 256x256),

% the restoration is an iterative scheme. this value defines maximal number
% of iteration s. For color images it is smaller (line 198).
ITER_MAX = 20;

% the colorspace transformation if image is color
COLOR_METHOD = 'OPP'; % 'OPP', 'RGB', 'DCT', 'YUV', 'PCA'

% true PSFs for 3 observations
Input.v_true{1} = function_GetPSF('average',9);
Input.v_true{2} = function_GetPSF('ggh_45');
Input.v_true{3} = function_GetPSF('inv_quad',7);
% Input.v_true{4} = function_GetPSF('motion',7,0);

% level of noise of all observations
Input.BSNR(1) = 40;
Input.BSNR(2) = 40;
Input.BSNR(3) = 40;
% Input.BSNR(4) = 40;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% Misalignment SHIFTS Settings: %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% images shifts:
Input.shifts(1,1) = 0; Input.shifts(1,2) = 0;
Input.shifts(2,1) = 0; Input.shifts(2,2) = 0;
Input.shifts(3,1) = 0; Input.shifts(3,2) = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Recursive gradient-projection algorithm settings:%%%%%%%%%%%%%%%%%%%
% normalize the image at every iteration to be in [0,1]
IMAGE_normalization = 'ON'; % 'ON', 'OFF'

% positivity of PSFs are assumed to be known a priory
PSF_positivity = 'ON'; % 'ON', 'OFF'

% PSFs are normalized on every iteration - 'ON'
PSF_normalization = 'ON'; % 'ON', 'OFF'

% number of additional iterations of PSFs without LPA-ICI filtering
AdditionalPSFIterations = 7; 

% step sizes in the iterative scheme
stepsize_v=.6; % for PSF reconstruction
stepsize_y=.9; % for the signal reconstruction

deltaPSF = 15; % SUPPORT size for PSF projection, odd sizes only
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Errors of results are shown of type:
error_type = 2; % 1 - ISNR(*) Improvement in Signal-to-Noise Ratio
                % 2 - SNR     Signal-to-Noise Ratio
                % 3 - PSNR    Peak Signal-to-Noise Ratio
                % 4 - MSE     Mean Squared Error          (l-2 norm squared)
                % 5 - RMSE    Root of Mean Squared Error  (l-2 norm)
                % 6 - MAE     Mean Absolute Error         (l-1 norm)
                % 7 - MAX     Maximum Absolute Difference (l-infinity norm)

% ----------------------------------------------------------------------
% Internal ALGORITHM PARAMETERS
% not recommended to be changed
% ----------------------------------------------------------------------

Input.ColorLength = 3; % number of color channels processed

% Weights in the criterion
lambda1=1.2;  % cross-term
lambda2=0.00001; % Y
lambda3=0.00001; % V_j

sst_filtering_y=0; % LPA-ICI filtering of SIGNAL begins after iteration 
                   % number sst_filtering_y
sst_filtering_v=0; % LPA-ICI filtering of PSFs begins after iteration 
                   % number sst_filtering_v


% ----------------------------------------------------------------------
% EXPERIMENT SETUP
% ----------------------------------------------------------------------
if strcmp(lower(EXPERIMENT),'fruits'),
    % true test color image
    Input.y = im2double(imread('image_Fruits512.png')); Input.y = Input.y(1:2:end,1:2:end,:); % Input.y = Input.y(32:32+127,32:32+127,:);
    % Input.y = im2double(imread('kodim05.png')); Input.y = Input.y(1:2:end,1:2:end,:);
    
    % the restoration is an iterative scheme. this value defines maximal number
    % of iteration s
    ITER_MAX = 15;
    if ~strcmp(lower(COLOR_METHOD),'rgb')
        GammaSignal = [.7 .9 1.];
    else
        GammaSignal = [.9 .9 .9];
    end;
    
elseif strcmp(lower(EXPERIMENT),'cameraman'),
    % true test grayscale image
    Input.y = im2double(imread('image_Cameraman256.png')); % Input.y=Input.y(12:31+108,61:60+128);    
    
elseif strcmp(lower(EXPERIMENT),'lena'),
    % true test grayscale image
    Input.y = im2double(imread('image_Lena256.png'));

elseif strcmp(lower(EXPERIMENT),'boats'),
    % true test grayscale image
    Input.y = im2double(imread('image_Boats512.png'));
    % the restoration is an iterative scheme. this value defines maximal number
    % of iteration s
    ITER_MAX = 15;
    
elseif strcmp(lower(EXPERIMENT),'text'),
    % true test grayscale image
    Input.y = im2double(imread('image_Text256.png'));
    
elseif strcmp(lower(EXPERIMENT),'testpat'),
    % true test grayscale image
    Input.y = im2double(imread('image_Testpat256.png'));
    
end;

% estimate variance of the noise (1) or use true value (0)?
estimate_sigma = 1;


% ----------------------------------------------------------------------
% RUN SIMULATIONS
% ----------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% INITIALIZATION %%%%%%%%%%%%%%%%%%%%%%%%%%

tic

v_support = zeros(deltaPSF + 2*max(abs(Input.shifts(:))),deltaPSF + 2*max(abs(Input.shifts(:))));
            % SUPPORT of PSFs, odd sizes only
            % used for errors, illustration of PSFs
            
size_initial=5; % support size of INITIAL PSF

delta11 = floor(deltaPSF/2) + 2*max(abs(Input.shifts(:)));
          % support size for PSF 
          % projection is in increased due to images shifts

% exclude border effect
border = 8;

% error check. number of parameters (BSNR(j) and shifts(j,1:2)) equal to a
% number of PSFs (v(j))
if length(Input.BSNR)<length(Input.v_true),
    Input.BSNR(length(Input.BSNR)+1:length(Input.v_true)) = 40; % default BSNR 40
end;
if size(Input.shifts,1)<length(Input.v_true),
    Input.shifts(size(Input.shifts,1)+1:length(Input.v_true),1:2)=0; % default shift 0
end;
          
Input.NumberOfObservations = min(length(Input.BSNR),length(Input.v_true));

disp(' ')
disp('Multi-Channel Blind Deconvolution...')
disp(' ')

% generate observations
y = function_MC_GenerateObservations(Input,size_initial,v_support);
Input.v_init = y.v_init;
Input.v_true = y.v_true;
Input.sigma  = y.sigma;
Input.z      = y.z;
Input.Z      = y.Z;

% number of processed color channels
Input.ColorLength = min(Input.ColorLength, size(Input.y,3));

for i=1:Input.NumberOfObservations;
    %%%%%%%%% COLORSPACE TRANSFORMATION %%%%%%%%%%%%%%%%%%%%%%
    if ~strcmp(lower(COLOR_METHOD),'rgb') & size(Input.y,3)>1,
        [a colormode] = function_rgb2LumChrom(squeeze(Input.z(:,:,i,:)),lower(COLOR_METHOD));
    else
        a = squeeze(Input.z(:,:,i,:));
    end;

    %%%%% observations and PSFs %%%%
    for ccolor=1:size(Input.y,3),
        Output.z(:,:,i,ccolor) = a(:,:,ccolor);
        Output.Z(:,:,i,ccolor) = fft2(Output.z(:,:,i,ccolor));

        % assume that PSF is not changed after possible color transformation
        Output.V(:,:,i,ccolor) = fft2(Input.v_init(:,:,i,ccolor));
        Output.v(1:size(v_support,1),1:size(v_support,2),i,ccolor) = Input.v_init(1:size(v_support,1),1:size(v_support,2),i,ccolor);

        Input.V_init(:,:,i,ccolor) = fft2(Input.v_init(:,:,i,ccolor));
    end;
end;

%%%% initial values for each color %%%%%%%%%
for ccolor=1:size(Input.y,3),
    Output.Y_est(:,:,ccolor) = sum(Output.Z(:,:,:,ccolor),3)./Input.NumberOfObservations;
    Output.y_est(:,:,ccolor) = real(ifft2(Output.Y_est(:,:,ccolor)));

    Output.y_initial(:,:,ccolor) = real(ifft2(Output.Y_est(:,:,ccolor)));
end;
%%%%%%%%%%%%%%%%% ~END OF INITIALIZATION %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  ESTIMATION OF THE SIGMA %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if estimate_sigma==0,
    Output.sigma = Input.sigma;
else    %%% estimate using MAD ('a-la-Donoho')
    for i=1:Input.NumberOfObservations,
        for ccolor=1:size(Input.y,3),
            sigma_est = function_stdEst2D(Output.z(:,:,i,ccolor),2);
            Output.sigma(ccolor,i) = sigma_est;
        end;
    end;
end
%%%%%%%%  ~END OF ESTIMATION OF THE SIGMA %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% ----------------------------------------------------------------------
% EXPERIMENT STARTS
% ----------------------------------------------------------------------
htimebar = timebar('MC Blind Deconvolution...','Progress'); itc=0;
totalcounter = ITER_MAX*Input.ColorLength;

yN = size(Input.y,1);
xN = size(Input.y,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% SHOW ITERATION INFORMATION %%%%%%%%%%%%%%%%%%%%%
h1 = figure; sst = 1;

for ccolor=1:Input.ColorLength,
    %%%%%%%%%%%%%%%%%% PSF Errors %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    for i=1:Input.NumberOfObservations,
        v = fftshift(real(ifft2(Output.V(:,:,i,ccolor))));
        v = function_KernelsForFFT(v, v_support);
        Output.v(1:size(v_support,1),1:size(v_support,2),i,ccolor) = v(1:size(v_support,1),1:size(v_support,2));

        % 'v Errors'
        in_v  = function_KernelsForFFT( Input.v_true{i}, zeros(yN,xN));
        out_v = fftshift(real(ifft2(Output.V(:,:,i,ccolor))));
        ini_v = fftshift(Input.v_init(:,:,i,ccolor));

        [Err_RI,labels] = function_Errors(in_v(border:end-border+1,border:end-border+1),out_v(border:end-border+1,border:end-border+1),ini_v(border:end-border+1,border:end-border+1),0);
        v_errors(sst,i,ccolor,:) = Err_RI;
        v_labels = labels;

        % series of PSF estimates
        v_est_series(:,:,sst,i,ccolor) = Output.v(:,:,i,ccolor);
    end;
end; % ~ccolor

% signal errors
a = Input.y(border:end-border+1,border:end-border+1,:);
if ~strcmp(lower(COLOR_METHOD),'rgb') & size(Input.y,3)>1,
    % colorspace transformation
    y_tmp = Output.y_est;
    if strcmp(lower(COLOR_METHOD),'pca'),
        y_tmp = LumChrom2rgb(Output.y_est,colormode);
        b = y_tmp(border:end-border+1,border:end-border+1,:);

        c = LumChrom2rgb(Output.y_initial,colormode);
        c = c(border:end-border+1,border:end-border+1,:);
    else
        y_tmp = function_LumChrom2rgb(Output.y_est,lower(COLOR_METHOD));
        b = y_tmp(border:end-border+1,border:end-border+1,:);

        c = function_LumChrom2rgb(Output.y_initial,lower(COLOR_METHOD));
        c = c(border:end-border+1,border:end-border+1,:);
    end;
else
    b = Output.y_est(border:end-border+1,border:end-border+1,:);
    c = Output.y_initial(border:end-border+1,border:end-border+1,:);
end;

for ccolor=1:size(Input.y,3),
    [Err_RI,labels] = function_Errors(a(:,:,ccolor),b(:,:,ccolor),c(:,:,ccolor),0); % 'y Errors vs y_est initial
    y_errors(sst,ccolor,:)=Err_RI;
    y_label=labels;
end;

y_tmp = Output.y_est;
% colorspace transformation for final estimate
if size(Input.y,3)>1,
    if strcmp(lower(COLOR_METHOD),'pca'),
        y_tmp = function_LumChrom2rgb(y_tmp,colormode);
    elseif ~strcmp(lower(COLOR_METHOD),'rgb'),
        y_tmp = function_LumChrom2rgb(y_tmp,lower(COLOR_METHOD));
    end;
    ttl = ['Iter. ',num2str(sst),'; ',v_labels(error_type,:),...
        'R: ',num2str(y_errors(sst,1,error_type),'%4.2f'),...
        ' ; G: ',num2str(y_errors(sst,2,error_type),'%4.2f'),...
        ' ; B: ',num2str(y_errors(sst,3,error_type),'%4.2f')];
else
    ttl = ['Iter. ',num2str(sst),'; ',v_labels(error_type,:),' ',num2str(y_errors(sst,1,error_type),'%4.2f')];
end;
y_est_series(:,:,sst,:) = y_tmp;

figure(h1),
subplot(1,2,1), imshow(y_tmp,[]), title(ttl);
if size(Input.y,3)>1,
    subplot(1,2,2), plot(y_errors(:,2,error_type)),
    title([v_labels(error_type,:),' Green Channel']),
else
    subplot(1,2,2), plot(y_errors(:,1,error_type)),
    title([v_labels(error_type,:)]);
end;
xlabel('Iterations'), xlim([1 ITER_MAX+1]);
axis square

tab_content = '';
if size(Input.y,3)>1,
    empty_array(1:7,1:5) = ' ';
    tab_content(1,:) = (['Est. errors: Iter.',num2str(sst),', Color trans. ',COLOR_METHOD]);
    tab_tmp = (['Criterion: R         G         B']);
    tab_content(2,1:length(tab_tmp)) = tab_tmp;
    tab_tmp = ([y_label empty_array num2str(squeeze(y_errors(end,1,:)),'%4.2f') ...
        empty_array  num2str(squeeze(y_errors(end,2,:)),'%4.2f') ...
        empty_array  num2str(squeeze(y_errors(end,3,:)),'%4.2f') ...
        empty_array  repmat('Wait...',[size(y_label,1) 1])]);
    tab_content(3:size(y_label,1)+2,1:size(tab_tmp,2)) = tab_tmp;
else
    empty_array(1:7,1:7) = ' ';
    tab_content(1,:) = ['Est. errors: Iter.',num2str(sst)];
    tab_tmp = [y_label empty_array num2str(squeeze(y_errors(sst,1,:)),'%4.2f'), empty_array, repmat('Wait...',[size(y_label,1) 1])];
    tab_content(2:size(y_label,1)+1,1:size(tab_tmp,2)) = tab_tmp;
end;

disp(tab_content);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for sst=1:ITER_MAX,
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% COLOR PROCESSING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    for ccolor=1:Input.ColorLength,
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        itc = itc+1;  timebar(htimebar,itc/totalcounter);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%% Invariant weights %%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        for i=1:Input.NumberOfObservations,
            for j=1:Input.NumberOfObservations,
                Output.d(i,j,ccolor) =...
                    sum(sum(Output.sigma(ccolor,j)^2.*abs(Output.V(:,:,i,ccolor)).^2+...
                    Output.sigma(ccolor,i)^2.*abs(Output.V(:,:,j,ccolor)).^2))/yN/xN;
            end;
            Output.d(i,i,ccolor) = Output.sigma(ccolor,i)^2;
        end;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%% Estimation of Signal  %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        tmp1 = 0;
        tmp2 = 0;
        for i=1:Input.NumberOfObservations,
            tmp1 = tmp1 + Output.Z(:,:,i,ccolor).*conj(Output.V(:,:,i,ccolor))/(Output.sigma(ccolor,i)^2);

            tmp2 = tmp2 + (abs(Output.V(:,:,i,ccolor)).^2)./(Output.sigma(ccolor,i)^2);
        end;
        Y_est2 = tmp1./(tmp2 + lambda2);

        Output.Y_est(:,:,ccolor) = Output.Y_est(:,:,ccolor)*(1-stepsize_y)+stepsize_y*Y_est2;

        % Filtering of signal estimate
        if sst>sst_filtering_y
            y_est = real(ifft2(Output.Y_est(:,:,ccolor)));
            dev2 = function_stdEst2D(y_est,2); y_var1(sst+1)=dev2;
            size_z_1=yN; size_z_2=xN;
            utility_MC_Signal_Filtering  % y_est_y result
        else
            y_est_y = real(ifft2(Output.Y_est(:,:,ccolor)));
        end;
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%% Signal Normalization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % normalize the image to be in [0,1] at every iteration
        if strcmp(lower(IMAGE_normalization),'on')
            y_est_y=max(0,min(y_est_y,1));
        end;
        
        Output.y_est(:,:,ccolor) = y_est_y;
        Output.Y_est(:,:,ccolor) = fft2(y_est_y);
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%% Estimation of PSF  %%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        for kk=1:AdditionalPSFIterations,
            for j=1:Input.NumberOfObservations,
                tmp1 = 0;
                tmp2 = 0;
                for i=1:Input.NumberOfObservations,
                    if (i < j) || (i > j),
                        tmp1 = tmp1 + conj(Output.Z(:,:,i,ccolor)).*Output.V(:,:,i,ccolor)./Output.d(j,i,ccolor);
                        tmp2 = tmp2 + abs(Output.Z(:,:,i,ccolor)).^2./Output.d(j,i,ccolor);
                    end;
                end;

                V_est_tmp(:,:,j,ccolor) =...
                    (Output.Z(:,:,j,ccolor).*conj(Output.Y_est(:,:,ccolor))/Output.sigma(ccolor,j)^2 + ...
                    lambda1.*Output.Z(:,:,j,ccolor).*tmp1)./...
                    (lambda3 + abs(Output.Y_est(:,:,ccolor)).^2/Output.sigma(ccolor,j)^2 + ...
                    lambda1.*tmp2);
            end;

            for j=1:Input.NumberOfObservations,
                Output.V(:,:,j,ccolor) = Output.V(:,:,j,ccolor)*(1-stepsize_v)+stepsize_v*V_est_tmp(:,:,j);
            end;
        end;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%% PSF Filtering %%%%%%%%%%%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if sst>sst_filtering_v,
            for j=1:Input.NumberOfObservations,
                %% fftshift is used in order to put PSF more or less in center of the support
                vvv = fftshift(real(ifft2(Output.V(:,:,j,ccolor))));
                
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                %%% PSF PROJECTION (Optional) %%%%%%
                aa1=ceil((yN+1)/2); aa2=ceil((xN+1)/2);
                vvv=vvv(aa1-delta11:aa1+delta11,aa2-delta11:aa2+delta11);
                % estimation of noise
                dev1 = function_stdEst2D(vvv,2); v_var1(sst+1)=dev1;
                utility_MC_PSF_Filtering % y_hat_PSF result
                vj_est=zeros([yN xN]);
                vj_est(aa1-delta11:aa1+delta11,aa2-delta11:aa2+delta11)=y_hat_PSF;

                % if positivity of PSFs are assumed to be known a priory
                if strcmp(lower(PSF_positivity),'on')
                    vj_est = vj_est.*double((vj_est>=0));
                end;
                
                % if PSFs are set to be normalized on every iteration
                if strcmp(lower(PSF_normalization),'on')
                    vj_est = vj_est/(sum(vj_est(:))+eps);
                end;

                Output.V(:,:,j,ccolor) = fft2(fftshift(vj_est));
                %%%%%%%% ~end of PSF PROJECTION %%%%
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            end;
        end;

        %%%%%%%%%%%%%%%%%% PSF Errors %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        for i=1:Input.NumberOfObservations,
            v = fftshift(real(ifft2(Output.V(:,:,i,ccolor))));
            v = function_KernelsForFFT(v, v_support);
            Output.v(1:size(v_support,1),1:size(v_support,2),i,ccolor) = v(1:size(v_support,1),1:size(v_support,2));

            % 'v Errors'
            in_v  = function_KernelsForFFT( Input.v_true{i}, zeros(yN,xN));
            out_v = fftshift(real(ifft2(Output.V(:,:,i,ccolor))));
            ini_v = fftshift(Input.v_init(:,:,i,ccolor));

            [Err_RI,labels] = function_Errors(in_v(border:end-border+1,border:end-border+1),out_v(border:end-border+1,border:end-border+1),ini_v(border:end-border+1,border:end-border+1),0);
            v_errors(sst+1,i,ccolor,:) = Err_RI;
            v_labels = labels;
            
            % series of PSF estimates
            v_est_series(:,:,sst+1,i,ccolor) = Output.v(:,:,i,ccolor);
        end;

    end; % ~ccolor
    %%%% ~END OF COLOR PROCESSING %%%%%%%%%%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %%%% SHOW ITERATION INFORMATION %%%%%%%%%%%%%%%%%%%%%
    % signal errors
    a = Input.y(border:end-border+1,border:end-border+1,:);
    if ~strcmp(lower(COLOR_METHOD),'rgb') & size(Input.y,3)>1,
        % colorspace transformation
        y_tmp = Output.y_est;
        if strcmp(lower(COLOR_METHOD),'pca'),
            y_tmp = LumChrom2rgb(Output.y_est,colormode);
            b = y_tmp(border:end-border+1,border:end-border+1,:);

            c = LumChrom2rgb(Output.y_initial,colormode);
            c = c(border:end-border+1,border:end-border+1,:);
        else
            y_tmp = function_LumChrom2rgb(Output.y_est,lower(COLOR_METHOD));
            b = y_tmp(border:end-border+1,border:end-border+1,:);

            c = function_LumChrom2rgb(Output.y_initial,lower(COLOR_METHOD));
            c = c(border:end-border+1,border:end-border+1,:);
        end;
    else
        b = Output.y_est(border:end-border+1,border:end-border+1,:);
        c = Output.y_initial(border:end-border+1,border:end-border+1,:);
    end;

    for ccolor=1:size(Input.y,3),
        [Err_RI,labels] = function_Errors(a(:,:,ccolor),b(:,:,ccolor),c(:,:,ccolor),0); % 'y Errors vs y_est initial
        y_errors(sst+1,ccolor,:)=Err_RI;
        y_label=labels;
    end;
    
    y_tmp = Output.y_est;
    % colorspace transformation for final estimate
    if size(Input.y,3)>1,        
        if strcmp(lower(COLOR_METHOD),'pca'),
            y_tmp = function_LumChrom2rgb(y_tmp,colormode);
        elseif ~strcmp(lower(COLOR_METHOD),'rgb'),
            y_tmp = function_LumChrom2rgb(y_tmp,lower(COLOR_METHOD));
        end;
        ttl = ['Iter. ',num2str(sst+1),'; ',v_labels(error_type,:),...
            'R: ',num2str(y_errors(sst+1,1,error_type),'%4.2f'),...
            ' ; G: ',num2str(y_errors(sst+1,2,error_type),'%4.2f'),...
            ' ; B: ',num2str(y_errors(sst+1,3,error_type),'%4.2f')];
    else
        ttl = ['Iter. ',num2str(sst+1),'; ',v_labels(error_type,:),' ',num2str(y_errors(sst+1,1,error_type),'%4.2f')];
    end;
    y_est_series(:,:,sst+1,:) = y_tmp;
    figure(h1), 
    subplot(1,2,1), imshow(y_tmp,[]), title(ttl);
    if size(Input.y,3)>1,
        subplot(1,2,2), plot(y_errors(:,2,error_type)), 
        title([v_labels(error_type,:),' Green Channel']),
    else
        subplot(1,2,2), plot(y_errors(:,1,error_type)), 
        title([v_labels(error_type,:)]);
    end;
    xlabel('Iterations'), xlim([1 ITER_MAX+1]);
    axis square
    drawnow

    if sst>0,
        disp(sprintf(repmat('\b',[1 numel(tab_content)+size(tab_content,1)+1]))) % clears previous table
    end;
    
    tab_content = '';
    if size(Input.y,3)>1,
        empty_array(1:7,1:5) = ' ';
        tab_content(1,:) = (['Est. errors: Iter.',num2str(sst+1),', Color trans. ',COLOR_METHOD]);
        tab_tmp = (['Criterion: R         G         B']);
        tab_content(2,1:length(tab_tmp)) = tab_tmp;
        tab_tmp = ([y_label empty_array num2str(squeeze(y_errors(end,1,:)),'%4.2f') ...
            empty_array  num2str(squeeze(y_errors(end,2,:)),'%4.2f') ...
            empty_array  num2str(squeeze(y_errors(end,3,:)),'%4.2f') ...
            empty_array  repmat('Wait...',[size(y_label,1) 1])]);
        tab_content(3:size(y_label,1)+2,1:size(tab_tmp,2)) = tab_tmp;
    else
        empty_array(1:7,1:7) = ' ';
        tab_content(1,:) = ['Est. errors: Iter.',num2str(sst+1)];
        tab_tmp = [y_label empty_array num2str(squeeze(y_errors(sst+1,1,:)),'%4.2f'), empty_array, repmat('Wait...',[size(y_label,1) 1])];
        tab_content(2:size(y_label,1)+1,1:size(tab_tmp,2)) = tab_tmp;
    end;

    disp(tab_content);
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
end % END of the sst iterations
disp(sprintf(repmat('\b',[1 numel(tab_content)+size(tab_content,1)+1]))) % clears previous table
close(h1);
close(htimebar);

% ----------------------------------------------------------------------
% Algorithm RESULTS
% ----------------------------------------------------------------------

disp(['time: ',num2str(toc/60)])
disp(' ')

% colorspace transformation for final estimate
if ~strcmp(lower(COLOR_METHOD),'rgb') & size(Input.y,3)>1,
    y_tmp = Output.y_est;
    if strcmp(lower(COLOR_METHOD),'pca'),
        Output.y_est = function_LumChrom2rgb(Output.y_est,colormode);
    else
        Output.y_est = function_LumChrom2rgb(Output.y_est,lower(COLOR_METHOD));
    end;
end;

% SHOW ESTIMATES
figure,
subplot(1,2,1), imshow(Input.y,[0 1]), title('True image');
subplot(1,2,2), imshow(Output.y_est,[0 1]), title('Estimate');

% SHOW PSFs
if size(Input.y,3)>1,
    figure,
    for i=1:size(Input.y,3),
        subplot(2,size(Input.y,3),i), imshow(Output.y_est(:,:,i),[0 1]), title(['y_{est}. ',y_label(error_type,:),': ',num2str(y_errors(end,i,error_type))]);
        subplot(2,size(Input.y,3),i+size(Input.y,3)), imshow(squeeze(Input.y(:,:,i)),[0 1]), title(['y_{true}. Channel: ',num2str(i)]);
    end;
end;

% SHOW OBSERVATIONS
figure, for i=1:Input.NumberOfObservations, subplot(2,Input.NumberOfObservations,i), imshow(squeeze(Input.z(:,:,i,:)),[0 1]), title(['Observation ',num2str(i)]), end;
subplot(2,Input.NumberOfObservations,Input.NumberOfObservations+ceil(Input.NumberOfObservations/2)), imshow(Output.y_est,[0 1]), title('Reconstruction');

% SHOW ONLY THE FIRST CHANNEL
for ccolor=1:1, % Input.ColorLength,
    % SHOW PSFs
    figure('name',['PSFs: Channel ',num2str(ccolor)]),
    for i=1:Input.NumberOfObservations,
        subplot(2,Input.NumberOfObservations,i), imshow(Input.v_true{i},[]), title(['True: v_',num2str(i)]),

        subplot(2,Input.NumberOfObservations,i+Input.NumberOfObservations), imshow(Output.v(:,:,i,ccolor),[]), title(['Estimated: v_',num2str(i)]),
    end;

    % SHOW PLOTS OF ERRORS
    figure('name',['ERROR: Channel ',num2str(ccolor)]),
    subplot(2,2,1), plot(y_errors(:,ccolor,error_type)), grid on, title('y_{est}'), ylabel(y_label(error_type,:)), xlabel('iterations'), xlim([1 ITER_MAX+1]);
    cc=1; j=1; cc1=1;
    for i=1:Input.NumberOfObservations,
        if cc1+i>4*cc, figure('name',['ERROR: Channel ',num2str(ccolor)]), cc=cc+1; j=0; cc1=0; end;
        subplot(2,2,1+j), plot(v_errors(:,i,ccolor,error_type)), grid on, title(['v_',num2str(i)]), ylabel(v_labels(error_type,:)), xlabel('iterations'), xlim([1 ITER_MAX+1]);
        j=j+1;
    end;
end;

    
% display final numerical results
if size(Input.y,3)>1,
    empty_array(1:7,1:5) = ' ';
    disp(['Estimate errors. Color trans: ','  ',COLOR_METHOD])
    disp(['Criterion: R         G         B'])
    disp([y_label empty_array num2str(squeeze(y_errors(end,1,:)),'%4.2f') empty_array  num2str(squeeze(y_errors(end,2,:)),'%4.2f') empty_array  num2str(squeeze(y_errors(end,3,:)),'%4.2f')])
else    
    disp(['Estimate errors:'])
    function_Errors(Input.y(border:end-border+1,border:end-border+1),...
        Output.y_est(border:end-border+1,border:end-border+1),...
        Output.y_initial(border:end-border+1,border:end-border+1),1);
end;
% ----------------------------- END -------------------------------------

